require 'octokit'

module BBFlow
  class Github
    class << self
      extend BBFlow::Persistent
      extend Memoist

      # @return [String]
      def http_url
        "https://github.com/#{repository_name}"
      end

      # @return [String]
      persistent local def repository_url
        remote.url
      end

      # @return [Array<Sawyer::Resource>]
      memoize def commits_info
        octokit.compare(repository_name, Options.get(:base_branch), local_branch).commits
      end

      # People who worked with you on this branch.
      #
      # @return [Array<String>]
      memoize def other_branch_committers
        (commits_info.map(&:author) + commits_info.map(&:committer)).map(&:login).uniq - [octokit.user.login]
      end

      # @return [Sawyer::Resource]
      def find_pull_request
        octokit.pull_requests(repository_name).detect { |pr| pr.head.ref == local_branch }
      end

      # @param [String] title
      # @param [String] body
      #
      # @return [Sawyer::Resource] the newly created pull request.
      def create_pull_request(title, body)
        puts 'Creating a pull request...'
        octokit.create_pull_request(repository_name, Options.get(:base_branch), local_branch, title, body)
      end

      # @param [String] pull_request
      # @param [Array<String>] reviewer_logins
      #
      # @return [void]
      def add_reviewers_to_pull_request(pull_request, reviewer_logins)
        puts 'Adding reviewers to the pull request...'
        octokit.request_pull_request_review(repository_name, pull_request, reviewer_logins)
      end

      # @return [Array<Array<String>>] Array<[login, name]>
      persistent local def collaborators
        puts 'Fetching project collaborators...'
        octokit.collaborators(repository_name).map(&:login).sort.map(&octokit.method(:user)).map { |user| [user.login, user.name] }
      end

      # @return [Array<String>]
      persistent local def labels
        puts 'Fetching repository labels...'
        octokit.labels(repository_name).map(&:name)
      end

      # @param [Sawyer::Resource] pull_request
      # @param [Array<String>] labels
      #
      # @return [void]
      def add_labels_to_pull_request(pull_request, labels)
        puts 'Adding labels to pull request...'
        octokit.add_labels_to_an_issue(repository_name, pull_request, labels)
      end

      # @return [Boolean]
      def has_branch?
        Misc.git.is_remote_branch?(local_branch)
      end

      # @return [Boolean]
      def pushed?
        unpushed_commits_number == 0
      end

      # @return [Integer]
      def unpushed_commits_number
        git_command = "git rev-list --count HEAD ^#{remote_branch}"
        Misc.execute(git_command).strip.to_i
      end

      memoize def changed_files
        Misc.execute("git diff --name-only #{Options.get(:base_branch)} #{remote_branch}").split
      end

      # @return [String]
      def remote_name
        remote.name
      end

      # @return [String]
      def remote_branch
        "#{remote_name}/#{local_branch}"
      end

      private

      # @return [Octokit::Client]
      memoize def octokit
        Octokit::Client.new(access_token: token)
      end

      # @return [String]
      def repository_name
        # Match both git@github.com:shockone/bb-flow-test.git and https://github.com/shockone/bb-flow-test
        repository_url[%r{github\.com[/:]([\w-]+/[\w-]+)(\.git|$)}, 1]
      end

      # @return [Git::Remote]
      memoize def remote
        remotes = Misc.git.remotes.select { |remote| remote.url.include?('github') }

        case remotes.size
        when 0
          Printer.panic("You don't have any remotes set up.")
        when 1
          remotes.first
        else
          Printer.select_item(
            remotes,
            'Which remote would you like to push to?',
            formatter: ->(remote) { [remote.name, remote.url] }
          )
        end
      end

      # @return [String]
      def local_branch
        Misc.git.current_branch
      end

      # @return [String]
      persistent global def token
        url = 'https://github.com/settings/tokens/new'
        Printer.ask("Paste a Github token. You can generate it on #{url} (leave default scopes): ")
      end
    end
  end
end
