require 'yaml/store'

module BBFlow
  # A bit of meta programming that allows to replace
  #
  #   def trello_key
  #     @trello_key ||= Config.global.transaction do |store|
  #       store['trello'] ||= {}
  #       store['trello']['key'] ||= Printer.ask('Go to https://trello.com/app-key and paste Key: ')
  #     end
  #   end
  #
  # with
  #
  #   class Trello
  #     extend Persistent
  #
  #     persistent global def key
  #       Printer.ask('Go to https://trello.com/app-key and paste Key: ')
  #     end
  #   end
  #
  # The method with be memoized both within and across executions.
  module Persistent
    # @param [Array<Config, Symbol>] args
    def persistent(args)
      store, method_name = args

      module_with_persistent_method = Module.new do
        define_method(method_name) do
          prefix = name.gsub(/^.*::/, '').downcase
          instance_variable_name = "@#{method_name}_memo"

          accessor = proc do |hash|
            hash[prefix] ||= {}
            hash[prefix][method_name.to_s] ||= super()
          end

          instance_variable_get(instance_variable_name) ||
          instance_variable_set(instance_variable_name, (store.instance_variable_get('@lock').locked? ? accessor.call(store) : store.transaction(&accessor)))
        end
      end

      prepend module_with_persistent_method
    end

    # @!method global(method_name)
    # @!method local(method_name)
    #   @param [Symbol] method_name
    #   @note intended to be used with #persistent
    #
    #   @return [Array<Config, Symbol>]
    %i(global local).each do |store_helper|
      define_method(store_helper) do |method_name|
        [Store.get(store_helper), method_name]
      end
    end

    class Store
      FILE_NAME = '.bb_flow'
      DIRS = { global: ENV['HOME'], local: Dir.pwd }

      # @param [Symbol] name
      #
      # @return [YAML::Store]
      def self.get(name)
        @memo ||= {}
        @memo[name] ||= YAML::Store.new(location(name))
      end

      def self.location(storage_type)
        fail "Unknown store: #{storage_type}" unless DIRS.keys.include?(storage_type)

        File.join(DIRS[storage_type], FILE_NAME)
      end
    end
  end
end
