module BBFlow
  class Updater
    class << self
      extend BBFlow::Persistent

      CHECK_INTERVAL = 60 * 60 * 24 * 7

      # @return [void]
      def ensure_up_to_date!
        return if recently_checked? || !update_available?

        if Printer.simple_question("An update from #{current_version} to #{latest_version} is available. Should I install it?")
          command = "gem install bb-flow -v '#{latest_version}'"
          puts "Executing `#{command}`..."
          `#{command}`

          exit Printer.success("Hooray, the new version has been installed! You can find the changes here: https://github.com/brightbytes/bb-flow/blob/master/CHANGELOG.md")
        end
      rescue StandardError => exception
        Printer.error("An error occurred while trying to self-update the gem: #{exception.message}")
      ensure
        update_last_check_unixtime!
      end

      private

      # @return [Boolean]
      def update_available?
        puts "Looking for updates..."
        latest_version > current_version
      end

      # @return [Gem::Version]
      def latest_version
        fetcher.search_for_dependency(dependency).first.map { |n, _| n.version }.first || Gem::Version.new('0')
      end

      # @return [Gem::Version]
      def current_version
        Gem::Version.new(BBFlow::VERSION)
      end

      # @return [Gem::SpecFetcher]
      def fetcher
        Gem::SpecFetcher.new(Gem::SourceList.from(['https://rubygems.org/']))
      end

      # @return [Gem::Dependency]
      def dependency
        Gem::Dependency.new('bb-flow')
      end

      # @return [Boolean]
      def recently_checked?
        last_check_unixtime + CHECK_INTERVAL > Time.now.to_i
      end

      # @return [Fixnum]
      persistent global def last_check_unixtime
        0
      end

      # @return [void]
      def update_last_check_unixtime!
        Persistent::Store.get(:global).transaction { |store| store['updater']['last_check_unixtime'] = Time.now.to_i }
      end
    end
  end
end
